#
#  Copyright (c) 2022 IBM Corp.
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

import logging
import re
import string
import sys
from typing import Dict, Iterable, Sequence

import pandas as pd

from label_sleuth.data_access.core.data_structs import DisplayFields, LABEL_POSITIVE, Label, TextElement


def get_element_group_by_texts(texts: Sequence[str], workspace_id, dataset_name, data_access, doc_uri=None) \
        -> Iterable[TextElement]:
    """
    The user may import a large number of labeled instances, and these will not necessarily be given with a text
    element uri. Thus, the goal here is to efficiently query for a group of text elements using only the texts, and
    optionally a *doc_uri* that the texts belong to.
    The order of the returned elements DOES NOT match the order of the texts given as input.
    """
    # TODO When we pass a doc_id, it is important not to remove duplicates. But is it really necessary to remove duplicates when we don't?
    remove_duplicates = False if doc_uri is not None else True
    regex = '|'.join(f'^{re.escape(t)}$' for t in texts)
    elements = data_access.get_text_elements(workspace_id=workspace_id, dataset_name=dataset_name,
                                             sample_size=sys.maxsize, query=regex, is_regex=True,
                                             document_uri=doc_uri, remove_duplicates=remove_duplicates)
    return elements['results']


def process_labels_dataframe(workspace_id, dataset_name, data_access, labels_df_to_import: pd.DataFrame) \
        -> Dict[str, Dict[str, Dict[str, Label]]]:
    logging.warning("Currently label metadata and label_type are ignored")
    # replace punctuation with underscores in category names
    punctuation = string.punctuation.replace("'", "") + string.whitespace
    labels_df_to_import[DisplayFields.category_name] = labels_df_to_import[DisplayFields.category_name].apply(str)
    labels_df_to_import[DisplayFields.category_name] = labels_df_to_import[DisplayFields.category_name].apply(
        lambda x: x.translate(x.maketrans(punctuation, '_' * len(punctuation))))
    # convert binary labels in dataframe to boolean
    positive_indicators = {LABEL_POSITIVE, str(LABEL_POSITIVE), str(LABEL_POSITIVE).upper(),
                           str(LABEL_POSITIVE).lower(), 1, "1"}
    labels_df_to_import[DisplayFields.label] = labels_df_to_import[DisplayFields.label].apply(
        lambda x: True if x in positive_indicators else False)

    category_to_uri_to_label = {}
    for category_name, category_df in labels_df_to_import.groupby(DisplayFields.category_name):
        # Here we group the texts by label, and optionally also by doc_id. The goal is to be able to efficiently
        # query for a group of elements by their texts - using *get_element_group_by_texts* - rather than querying
        # and assigning a label for one text at a time.
        if DisplayFields.doc_id not in category_df.columns:
            doc_id_to_label_to_texts = {None:
                                            {label: df_for_label[DisplayFields.text].values.tolist()
                                             for label, df_for_label in category_df.groupby(DisplayFields.label)}}
        else:
            doc_id_to_label_to_texts = {doc_id:
                                            {label: df_for_label[DisplayFields.text].values.tolist()
                                             for label, df_for_label in df_for_doc.groupby(DisplayFields.label)}
                                        for doc_id, df_for_doc in category_df.groupby(DisplayFields.doc_id)}

        uri_to_label = {}
        for doc_id, label_to_texts in doc_id_to_label_to_texts.items():
            doc_uri = f'{dataset_name}-{doc_id}' if doc_id is not None else None
            for label, texts in label_to_texts.items():
                elements_to_label = get_element_group_by_texts(texts, workspace_id, dataset_name, data_access,
                                                               doc_uri=doc_uri)
                uri_to_label.update({e.uri: {category_name: Label(label=label)} for e in elements_to_label})
        category_to_uri_to_label[category_name] = uri_to_label
    return category_to_uri_to_label



#
#  Copyright (c) 2022 IBM Corp.
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

from label_sleuth.data_access.data_access_api import DataAccessApi
from label_sleuth.training_set_selector.train_set_selectors import TrainSetSelectorAllLabeled, \
    TrainSetSelectorEnforcePositiveNegativeRatio
from label_sleuth.training_set_selector.train_set_selector_api import TrainingSetSelectionStrategy


def get_training_set_selector(data_access: DataAccessApi, strategy=TrainingSetSelectionStrategy.ALL_LABELED):
    if strategy == TrainingSetSelectionStrategy.ALL_LABELED:
        return TrainSetSelectorAllLabeled(data_access)
    elif strategy == TrainingSetSelectionStrategy.ALL_LABELED_PLUS_UNLABELED_AS_NEGATIVE_EQUAL_RATIO:
        return TrainSetSelectorEnforcePositiveNegativeRatio(data_access, required_negative_ratio=1)
    elif strategy == TrainingSetSelectionStrategy.ALL_LABELED_PLUS_UNLABELED_AS_NEGATIVE_X2_RATIO:
        return TrainSetSelectorEnforcePositiveNegativeRatio(data_access, required_negative_ratio=2)
    elif strategy == TrainingSetSelectionStrategy.ALL_LABELED_PLUS_UNLABELED_AS_NEGATIVE_X10_RATIO:
        return TrainSetSelectorEnforcePositiveNegativeRatio(data_access, required_negative_ratio=10)

    else:
        raise Exception(f"{strategy} is not supported")



#
#  Copyright (c) 2022 IBM Corp.
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

from dataclasses import dataclass
from typing import Sequence


@dataclass
class Language:
    """
    This dataclass contains some parameters that are language-specific. The parameter values can be used by the
    classification model implementations, as well as by some general utility functions in the *analysis_utils* module.
    """
    name: str
    stop_words: Sequence[str]
    spacy_model_name: str

    def __repr__(self):
        return self.name


English = \
    Language(name='English',
             stop_words=["a", "about", "above", "after", "again", "against", "ain", "all", "am", "an", "and", "any",
                         "are", "aren", "aren't", "as", "at", "be", "because", "been", "before", "being", "below",
                         "between", "both", "but", "by", "can", "couldn", "couldn't", "d", "did", "didn", "didn't",
                         "do", "does", "doesn", "doesn't", "doing", "don", "don't", "down", "during", "each", "few",
                         "for", "from", "further", "had", "hadn", "hadn't", "has", "hasn", "hasn't", "have", "haven",
                         "haven't", "having", "he", "her", "here", "hers", "herself", "him", "himself", "his", "how",
                         "i", "if", "in", "into", "is", "isn", "isn't", "it", "it's", "its", "itself", "just", "ll",
                         "m", "ma", "me", "mightn", "mightn't", "more", "most", "mustn", "mustn't", "my", "myself",
                         "needn", "needn't", "no", "nor", "not", "now", "o", "of", "off", "on", "once", "only", "or",
                         "other", "our", "ours", "ourselves", "out", "over", "own", "re", "s", "same", "shan", "shan't",
                         "she", "she's", "should", "should've", "shouldn", "shouldn't", "so", "some", "such", "t",
                         "than", "that", "that'll", "the", "their", "theirs", "them", "themselves", "then", "there",
                         "these", "they", "this", "those", "through", "to", "too", "under", "until", "up", "ve", "very",
                         "was", "wasn", "wasn't", "we", "were", "weren", "weren't", "what", "when", "where", "which",
                         "while", "who", "whom", "why", "will", "with", "won", "won't", "wouldn", "wouldn't", "y",
                         "you", "you'd", "you'll", "you're", "you've", "your", "yours", "yourself", "yourselves",
                         "could", "he'd", "he'll", "he's", "here's", "how's", "i'd", "i'll", "i'm", "i've", "let's",
                         "ought", "she'd", "she'll", "that's", "there's", "they'd", "they'll", "they're", "they've",
                         "we'd", "we'll", "we're", "we've", "what's", "when's", "where's", "who's", "why's", "would"],
             spacy_model_name='en_core_web_lg')

Italiano = \
    Language(name='Italiano',
             stop_words=["scola", "varie", "saro", "stanno", "dagli", "certa", "le", "nemmeno", "molto", "avuti", 
							"posteriore", "facesse", "il", "nostra", "durante", "faceste", "dalla", "può", "della", "ora", "faccia", 
							"alla", "dice", "scorso", "probabilmente", "volta", "sareste", "tre", "fu", "dell", "quanta", "soltanto", 
							"fa", "degl", "diventare", "pure", "quantunque", "che", "anche", "pieno", "nostri", "te", "facessero", 
							"qualunque", "codesti", "nonsia", "come", "successivamente", "stato", "mesi", "quelle", "insieme", "agli", 
							"volte", "faresti", "nell", "lui", "ma", "malgrado", "ossia", "allo", "lato", "mi", "aveste", "lo", "molti", 
							"basta", "seguito", "dallo", "cioe", "giacche", "avevate", "starebbero", "parecchio", "ansa", "paese", 
							"partendo", "secondo", "prima", "gliela", "quale", "là", "uno", "vari", "facciamo", "dove", "sarà", "stati", 
							"medesimo", "steste", "male", "stiate", "facciano", "ahime", "dovrà", "tra", "stessero", "osi", "esse", 
							"seguente", "stavo", "percio", "sarebbe", "avrai", "parecchie", "codesto", "dall", "cominci", "tempo", "sui", 
							"relativo", "faremo", "aver", "staremo", "citta", "dei", "improvviso", "stesse", "verso", "generale", "po", 
							"avreste", "mie", "parecchi", "fuori", "uguali", "gruppo", "queste", "contro", "diventato", "col", "così", 
							"puo", "sarete", "quindi", "anni", "si", "stavamo", "mediante", "eravamo", "state", "stettero", "sto", 
							"colui", "facesti", "lungo", "altro", "fareste", "fummo", "agl", "solito", "certe", "abbia", "ai", "faremmo", 
							"stavano", "perfino", "no", "quelli", "certo", "miei", "feci", "fossi", "presa", "purtroppo", "con", 
							"facevi", "chi", "momento", "concernente", "stiamo", "ecc", "saremo", "successivo", "una", "per", "coloro", 
							"governo", "nessuno", "dappertutto", "conclusione", "fra", "farà", "glieli", "sotto", "salvo", "dentro", 
							"altri", "sarebbero", "fece", "tutto", "quel", "effettivamente", "luogo", "alle", "essi", "ti", "ella", 
							"sembrare", "casa", "foste", "fatto", "sarò", "starò", "la", "avuta", "tu", "codesta", "in", "sugl", 
							"staresti", "starà", "fanno", "starete", "mila", "degli", "stesti", "avute", "voi", "esser", "finalmente", 
							"avemmo", "ebbe", "peggio", "cosa", "nondimeno", "maggior", "lei", "grande", "bene", "avevo", "ore", "stava", 
							"nello", "stiano", "allora", "farò", "infatti", "sembrato", "dirimpetto", "nella", "me", "ero", "piglia", 
							"altrove", "nel", "saremmo", "ho", "entrambi", "ognuno", "altrimenti", "tanto", "avranno", "dal", 
							"recentemente", "tali", "sulle", "tua", "dovunque", "farebbe", "sei", "avessimo", "pero", "avremmo", "coll", 
							"facevate", "spesso", "avevano", "cio", "niente", "inoltre", "ancora", "otto", "li", "abbastanza", "varia", 
							"qualcuno", "vale", "già", "ieri", "vostra", "cento", "facemmo", "stette", "faceva", "persone", "avevamo", 
							"ci", "troppo", "però", "dunque", "sulla", "avrete", "all", "vi", "piuttosto", "saresti", "stessi", 
							"ottanta", "bravo", "sia", "ed", "proprio", "intanto", "quello", "nelle", "chiunque", "facevano", "visto", 
							"colei", "miliardi", "avrei", "press", "invece", "nulla", "ahimè", "altrui", "se", "sono", "sua", "gliele", 
							"consiglio", "ulteriore", "peccato", "recente", "tuoi", "lavoro", "frattempo", "avrà", "qualcosa", "eppure", 
							"sopra", "stando", "forza", "fossero", "attraverso", "nonostante", "gliene", "dai", "negli", "farete", 
							"moltissimo", "poiche", "sempre", "quante", "noi", "sembri", "dietro", "marche", "poco", "loro", "nei", 
							"vostre", "giorno", "ad", "città", "deve", "novanta", "oltre", "eri", "grazie", "milioni", "quest", 
							"talvolta", "cui", "glielo", "dalle", "lasciato", "certi", "facessi", "solo", "avrebbero", "essendo", 
							"quanti", "stia", "hai", "stemmo", "via", "dov", "staremmo", "avente", "nove", "perciò", "quella", "mai", 
							"tuo", "nostre", "lontano", "ex", "nostro", "ciascuno", "fino", "perché", "ultimo", "faranno", "ha", "inc", 
							"perche", "al", "ognuna", "sara", "modo", "ciascuna", "faccio", "qui", "chicchessia", "cos", "mentre", 
							"adesso", "sig", "hanno", "cosi", "avrò", "persino", "dovra", "avresti", "facendo", "favore", "benissimo", 
							"stareste", "sul", "da", "stata", "co", "quanto", "tale", "facessimo", "avanti", "dagl", "tuttavia", "farai", 
							"quando", "avendo", "preferibilmente", "avremo", "starebbe", "finche", "caso", "stetti", "malissimo", 
							"essere", "giorni", "stesso", "vostri", "vario", "trenta", "ecco", "fosti", "stai", "ne", "mia", "furono", 
							"vita", "sarei", "poi", "anticipo", "delle", "tranne", "principalmente", "coi", "stavate", "potrebbe", 
							"farei", "siete", "brava", "fin", "cima", "attesa", "stavi", "del", "fecero", "avesse", "forse", "detto", 
							"gia", "farebbero", "avessero", "saranno", "minimi", "piu", "senza", "starei", "sarai", "possedere", "non", 
							"pochissimo", "un", "primo", "piedi", "davanti", "titolo", "od", "uomo", "stessimo", "mio", "negl", "tutte", 
							"avrebbe", "suo", "fossimo", "fai", "io", "mosto", "parte", "facevo", "questa", "facevamo", "tutta", "sette", 
							"gli", "fui", "scopo", "vostro", "fine", "fosse", "trovato", "nessuna", "qualcuna", "riecco", "cogli", "sta", 
							"possa", "ebbero", "su", "avete", "fare", "nuovo", "posto", "a", "realmente", "meno", "registrazione", 
							"quasi", "mondo", "sullo", "comunque", "siano", "dopo", "esempio", "avevi", "siamo", "avesti", "nessun", 
							"tue", "va", "cortesia", "intorno", "questi", "ministro", "anno", "alcuni", "sugli", "tutti", "srl", 
							"diventa", "subito", "nazionale", "siate", "ebbi", "sue", "alcuna", "abbiate", "mezzo", "due", "meglio", 
							"eravate", "ogni", "macche", "dello", "magari", "più", "alcuno", "questo", "avere", "oggi", "vicino", 
							"futuro", "circa", "staranno", "stessa", "di", "avuto", "haha", "assai", "starai", "abbiamo", "era", 
							"oppure", "quali", "accidenti", "abbiano", "suoi", "egli", "neppure", "averlo", "qualche", "sull", "avessi", 
							"quattro", "facciate", "dire", "mancanza", "affinche", "erano", "conciliarsi", "sembra", "aveva"],
             spacy_model_name='it_core_news_lg')


class Languages:
    ENGLISH = English
    ITALIANO = Italiano



#
#  Copyright (c) 2022 IBM Corp.
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

import abc

from enum import Enum
from typing import Sequence

from label_sleuth.data_access.core.data_structs import TextElement


class TrainingSetSelectionStrategy(Enum):
    """
    Given the current set of elements labeled by the user, a TrainingSetSelectionStrategy determines which examples
    will be sent in practice to the model at training time. For example, the strategy may specify that additional
    _unlabeled_ elements will be given to the model as weak negative examples.
    """
    ALL_LABELED = 0
    ALL_LABELED_PLUS_UNLABELED_AS_NEGATIVE_EQUAL_RATIO = 1
    ALL_LABELED_PLUS_UNLABELED_AS_NEGATIVE_X2_RATIO = 2
    ALL_LABELED_PLUS_UNLABELED_AS_NEGATIVE_X10_RATIO = 3


class TrainSetSelectorAPI(object, metaclass=abc.ABCMeta):

    def __init__(self, data_access):
        self.data_access = data_access

    @abc.abstractmethod
    def get_train_set(self, workspace_id: str, train_dataset_name: str, category_id: int) -> Sequence[TextElement]:
        """
        For a given workspace, dataset and category, prepare and return a train set for training the model.
        Returns a list of TextElement objects (containing labels for the category, and possibly metadata about
        some of these labels).

        :param workspace_id:
        :param train_dataset_name:
        :param category_id:
        """
